// CAN2515.h
//
// CAN MCP2515ドライバ (for Arduino)
//

#include <avr/io.h>

#include "P2515Tag.h"
#include "P2515Reg.h"
#include "P2515SpiPort.h"		// SPIポートのI/Oアサイン

typedef unsigned char byte;
typedef unsigned int word;

// SPIポート ビット操作
#define CAN_SPI_CS_H	CAN_SPI_CS_PORT|=(1<<CAN_SPI_CS_BIT)
#define CAN_SPI_SO_H	CAN_SPI_SO_PORT|=(1<<CAN_SPI_SO_BIT)
#define CAN_SPI_SCK_H	CAN_SPI_SCK_PORT|=(1<<CAN_SPI_SCK_BIT)

#define CAN_SPI_CS_L	CAN_SPI_CS_PORT&=~(1<<CAN_SPI_CS_BIT)
#define CAN_SPI_SO_L	CAN_SPI_SO_PORT&=~(1<<CAN_SPI_SO_BIT)
#define CAN_SPI_SCK_L	CAN_SPI_SCK_PORT&=~(1<<CAN_SPI_SCK_BIT)

#define CAN_SPI_SI		(CAN_SPI_SI_PORT&(1<<CAN_SPI_SI_BIT))


// CANビットレート ロースピード 125KHz
#define CAN_BRP_20MHz_125KBPS	4			// FOSC=20MHz
//#define CAN_BRP_16MHz_125KBPS	3			// FOSC=16MHz
#define CAN_BRP_8MHz_125KBPS	2			// FOSC=8MHz

// CANビットレート X'tal 16MHz
#define CAN_BRP_16MHz_500KBPS	0			// FOSC=16MHz
#define CAN_BRP_16MHz_250KBPS	1			// FOSC=16MHz
#define CAN_BRP_16MHz_125KBPS	3			// FOSC=16MHz


// SPIコマンド インストラクション
#define SPI_INST_RESET		0xC0
#define SPI_INST_READ		0x03
#define SPI_INST_WRITE		0x02
#define SPI_INST_RD_RXBUF	0xC0
#define SPI_INST_LD_TXBUF	0x40
#define SPI_INST_RTS		0x80
#define SPI_INST_RD_STAT	0xA0
#define SPI_INST_RX_STAT	0xB0
#define SPI_INST_BIT_MOD	0x05

// 送／受信バッファ シーケンシャル・リード／ライト インストラクション
#define RD_RXBUF_RXB0SIDH	0x90
#define RD_RXBUF_RXB0D0		0x92
#define RD_RXBUF_RXB1SIDH	0x94
#define RD_RXBUF_RXB1D0		0x96
#define LD_TXBUF_TXB0SIDH	0x40
#define LD_TXBUF_TXB0D0		0x41
#define LD_TXBUF_TXB1SIDH	0x42
#define LD_TXBUF_TXB1D0		0x43
#define LD_TXBUF_TXB2SIDH	0x44
#define LD_TXBUF_TXB2D0		0x45

// *********** CAN ドライバ（マクロ） ************

// CANオペレーション・モード設定
#define CANSetOpMode(mode) CANBitModCmd(CANCTRL, 7<<5, mode<<5)
// オペレーション・モード CANCTRL/CANSTAT
#define CAM_MODE_NORMAL		0
#define CAM_MODE_SLEEP		1
#define CAM_MODE_LOOPBACK	2
#define CAM_MODE_LISTEN		3
#define CAM_MODE_CONFIG		4


// レジスタ バイト・リード
#define CANReadReg(adrs) CANRegRead2B(SPI_INST_READ, adrs)

// ステータス・リード
#define CANReadStat() CANRegRead2B(SPI_INST_RD_STAT, 0)

// 受信ステータス・リード
#define CANReadRXStat() CANRegRead2B(SPI_INST_RX_STAT, 0)

// フィルタ・モードの設定
#define CANSetFilterRxB0(mode) CANWriteReg(RXB0CTRL, mode<<5)
#define CANSetFilterRxB1(mode) CANWriteReg(RXB1CTRL, mode<<5)
// mode
#define FILTER_MOD_ALL_HIT 3		// マスク、フィルタ使用せず
#define FILTER_MOD_SID_HIT 1		// SIDでマスク、フィルタ適用


// 受信バッファ シーケンシャル・リード
#define CANRxB0MsgRead(buf) CANTxRxBufRW(RD_RXBUF_RXB0SIDH, 1, buf, 5)
#define CANRxB1MsgRead(buf) CANTxRxBufRW(RD_RXBUF_RXB1SIDH, 1, buf, 5)
#define CANRxB0DatRead(buf) CANTxRxBufRW(RD_RXBUF_RXB0D0, 1, buf, 8)
#define CANRxB1DatRead(buf) CANTxRxBufRW(RD_RXBUF_RXB1D0, 1, buf, 8)

// 受信バッファ シーケンシャル・リード(メッセージ＋データ)
#define CANRxB0Read(buf, dsiz) CANTxRxBufRW(RD_RXBUF_RXB0SIDH, 1, buf, 5+dsiz)
#define CANRxB1Read(buf, dsiz) CANTxRxBufRW(RD_RXBUF_RXB1SIDH, 1, buf, 5+dsiz)


// 送信バッファ シーケンシャル・ライト
#define CANTxB0MsgWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB0SIDH, 0, buf, 5)
#define CANTxB1MsgWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB1SIDH, 0, buf, 5)
#define CANTxB2MsgWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB2SIDH, 0, buf, 5)
#define CANTxB0DatWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB0D0, 0, buf, 8)
#define CANTxB1DatWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB1D0, 0, buf, 8)
#define CANTxB2DatWrite(buf) CANTxRxBufRW(LD_TXBUF_TXB2D0, 0, buf, 8)

// 送信バッファ シーケンシャル・ライト(メッセージ＋データ)
#define CANTxB0Write(buf, dsiz) CANTxRxBufRW(LD_TXBUF_TXB0SIDH, 0, buf, 5+dsiz)
#define CANTxB1Write(buf, dsiz) CANTxRxBufRW(LD_TXBUF_TXB1SIDH, 0, buf, 5+dsiz)
#define CANTxB2Write(buf, dsiz) CANTxRxBufRW(LD_TXBUF_TXB2SIDH, 0, buf, 5+dsiz)


// フィルタSID設定
#define CANSetSidFilter0(sid) CANSetSidFilteMask(RXF0SIDH, sid)
#define CANSetSidFilter1(sid) CANSetSidFilteMask(RXF1SIDH, sid)
#define CANSetSidFilter2(sid) CANSetSidFilteMask(RXF2SIDH, sid)
#define CANSetSidFilter3(sid) CANSetSidFilteMask(RXF3SIDH, sid)
#define CANSetSidFilter4(sid) CANSetSidFilteMask(RXF4SIDH, sid)
#define CANSetSidFilter5(sid) CANSetSidFilteMask(RXF5SIDH, sid)

// マスクSID設定
#define CANSetSidMask0(sid) CANSetSidFilteMask(RXM0SIDH, sid)
#define CANSetSidMask1(sid) CANSetSidFilteMask(RXM1SIDH, sid)

#define MASK_SID_ALL_HIT   0x0000		// すべてマスク
#define MASK_SID_CPL_MATCH 0x07FF		// マスク無効 完全一致

// SIDレジスタ値操作マクロ
#define GET_SIDH(msg) (byte)(msg >> 3)
#define GET_SIDL(msg) (byte)(msg << 5)
#define MAKE_SID(sid_h, sid_l) ((unsigned int)sid_h<<3)|(sid_l>>5)


// SPIコマンド 受信ステータス構造(受信ステータス・リード・コマンド)
struct tag_RXSTAT {
	unsigned char FilterMatch:3;		// b0-b2 ヒットしたフィルタの番号
	unsigned char RemoteFrm:1;		// b3 (1)リモート・フレーム/(0)データ・フレーム
	unsigned char ExtMsg:1;			// b4 (1)拡張ID/(0)標準ID
	unsigned char b5:1;				// b5
	unsigned char MsgInRXB0:1;		// b6 RXB0にメッセージあり
	unsigned char MsgInRXB1:1;		// b7 RXB1にメッセージあり
};


// プロトタイプ
void CANInit(byte brp);
byte CANRxCheck(byte rxbnum);
void CANReset(void);
void CANWriteReg(byte adrs, byte data);
byte CANRegRead2B(byte inst, byte adrs);
void CANSetSidFilteMask(byte adrs, word sid);
void CANBitModCmd(byte adrs, byte mask, byte data);
void CANTxRxBufRW(byte inst, byte readop, byte *data, byte cnt);


// --------------------------------------------------------------------
// MCP2515初期化
// --------------------------------------------------------------------
void CANInit(byte brp) {
//	struct tag_CNF1 *p_cnf1;
//	struct tag_CNF2 *p_cnf2;
//	struct tag_CNF3 *p_cnf3;
	byte reg;

	// SPIポートの初期化
	CAN_SPI_CS_DDR |= (1<<CAN_SPI_CS_BIT);		// output
	CAN_SPI_SO_DDR |= (1<<CAN_SPI_SO_BIT);		// output
	CAN_SPI_SCK_DDR |= (1<<CAN_SPI_SCK_BIT);	// output
	CAN_SPI_SI_DDR &= ~(1<<CAN_SPI_SI_BIT);		// input

	// SPIバス アイドル
	CAN_SPI_CS_H;
	CAN_SPI_SCK_L;

//	delay(10);
	CANReset();

//	delay(10);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), (1<<B0BFS) | (1<<B0BFE));		// LED3 ON
//	CANWriteReg(BFPCTRL, (1<<B0BFS) | (1<<B0BFE));
//	CANWriteReg(BFPCTRL, 0x14);

	CANSetOpMode(CAM_MODE_CONFIG);
	// MCP2515 コンフィギュレーション設定
	// ボーレート設定 125kbps (X=20MHz)
	// Tq = (2 x (BRP + 1))/Fosc )
//	reg = 0;
//	p_cnf1 = (struct tag_CNF1 *)&reg;
//	p_cnf1->BRP = brp;					// 1/10
//	p_cnf1->SJW0 = 0;
//	p_cnf1->SJW1 = 0;
reg = brp;
	CANWriteReg(CAN_CNF1, reg);

//	reg = 0;
//	p_cnf2 = (struct tag_CNF2 *)&reg;
//	p_cnf2->PRSEG = 1;					// ProgSeg = 2
//	p_cnf2->PHSEG1 = 6;				// PH1 = 7
//	p_cnf2->SAM = 0;
//	p_cnf2->BTLMODE = 1;
reg = 0xB1;
	CANWriteReg(CAN_CNF2, reg);

//	reg = 0;
//	p_cnf3 = (struct tag_CNF3 *)&reg;
//	p_cnf3->PHSEG2 = 5;				// PH2 = 6
//	p_cnf3->WAKFIL = 0;
//	p_cnf3->SOF = 0;
reg = 0x05;
	CANWriteReg(CAN_CNF3, reg);

//		CANSetOpMode(CAM_MODE_NORMAL);

}


// --------------------------------------------------------------------
// CAN受信チェック
// --------------------------------------------------------------------
byte CANRxCheck(byte rxbnum) {
	byte rxstat;

	rxstat = CANReadRXStat();			// 受信ステータス取得

	if(rxbnum == 0) {
		return rxstat & 1<<MsgInRXB0;
	} else if(rxbnum == 1) {
		return rxstat & 1<<MsgInRXB1;
	}

	return rxstat & ( (1<<MsgInRXB0) | (1<<MsgInRXB1) );
}


// --------------------------------------------------------------------
// CANリセット
// --------------------------------------------------------------------
void CANReset(void) {
	byte i;
	byte cmd = SPI_INST_RESET;

	CAN_SPI_CS_L;

	for(i = 0; i < 8; i++) {
		if(cmd & 0x80) {
			// 1
			CAN_SPI_SO_H;
		} else {
			// 0
			CAN_SPI_SO_L;
		}
		cmd <<= 1;
		CAN_SPI_SCK_H;		// SCK パルス出力
		CAN_SPI_SCK_L;
	}

	CAN_SPI_CS_H;
}


// --------------------------------------------------------------------
// CANレジスタ バイト・ライト
// --------------------------------------------------------------------
void CANWriteReg(byte adrs, byte data) {

	byte cmdbf[3], cmd;
	byte p, i;

	cmdbf[0] = SPI_INST_WRITE;		// バイト・ライト
	cmdbf[1] = adrs;
	cmdbf[2] = data;

	CAN_SPI_CS_L;

	for(p = 0; p < 3; p++) {
		cmd = cmdbf[p];
		for(i = 0; i < 8; i++) {
			if(cmd & 0x80) {
				// 1
				CAN_SPI_SO_H;
			} else {
				// 0
				CAN_SPI_SO_L;
			}
			cmd <<= 1;
			CAN_SPI_SCK_H;		// SCK パルス出力
			CAN_SPI_SCK_L;
		}
	}

	CAN_SPI_CS_H;
}


// --------------------------------------------------------------------
// CANレジスタ・リード
// ステータス・リード
// 受信ステータス・リード  (3機能兼用ルーチン)
// --------------------------------------------------------------------
byte CANRegRead2B(byte inst, byte adrs) {
	byte cmdbf[2], cmd;
	byte p, i, dat;

	cmdbf[0] = inst;		// インストラクション
	cmdbf[1] = adrs;

	CAN_SPI_CS_L;

	// 2バイト出力(データ入力は無視)
	for(p = 0; p < 2; p++) {
		cmd = cmdbf[p];
		for(i = 0; i < 8; i++) {
			if(cmd & 0x80) {
				// 1
				CAN_SPI_SO_H;
			} else {
				// 0
				CAN_SPI_SO_L;
			}
			cmd <<= 1;
			CAN_SPI_SCK_H;		// SCK パルス出力
			CAN_SPI_SCK_L;
		}
	}

	// データ入力(ステータスの場合は、2回目のリピートデータ)
	dat = 0;
	for(i = 0; i < 8; i++) {
		dat <<= 1;
		if(CAN_SPI_SI) {
			// 1
			dat |= 1;
		}
		CAN_SPI_SCK_H;		// SCK パルス出力
		CAN_SPI_SCK_L;
	}

	CAN_SPI_CS_H;

	return dat;
}


// --------------------------------------------------------------------
// CANビット・モデファイ・コマンド
// --------------------------------------------------------------------
void CANBitModCmd(byte adrs, byte mask, byte data) {
	byte cmdbf[4], cmd;
	byte p, i;

	cmdbf[0] = SPI_INST_BIT_MOD;		// ビット・モデファイ
	cmdbf[1] = adrs;
	cmdbf[2] = mask;
	cmdbf[3] = data;

	CAN_SPI_CS_L;

	for(p = 0; p < 4; p++) {
		cmd = cmdbf[p];
		for(i = 0; i < 8; i++) {
			if(cmd & 0x80) {
				// 1
				CAN_SPI_SO_H;
			} else {
				// 0
				CAN_SPI_SO_L;
			}
			cmd <<= 1;
			CAN_SPI_SCK_H;		// SCK パルス出力
			CAN_SPI_SCK_L;
		}
	}
	CAN_SPI_CS_H;
}


// --------------------------------------------------------------------
// CAN送／受信バッファ シーケンシャル・リード／ライト (兼用ルーチン)
// --------------------------------------------------------------------
void CANTxRxBufRW(byte inst, byte readop, byte *data, byte cnt) {

	byte dat;
	byte p, i;

	CAN_SPI_CS_L;

	// インストラクションの送信
	dat = inst;
	for(i = 0; i < 8; i++) {
		if(dat & 0x80) {
			// 1
			CAN_SPI_SO_H;
		} else {
			// 0
			CAN_SPI_SO_L;
		}
		dat <<= 1;
		CAN_SPI_SCK_H;		// SCK パルス出力
		CAN_SPI_SCK_L;
	}

	// =============== 送信/受信バッファ内容の送受信 =================
	if(readop) {
		// データ受信
		for(p = 0; p < cnt; p++) {
			dat = 0;
			for(i = 0; i < 8; i++) {
				dat <<= 1;
				if(CAN_SPI_SI) {
					// 1
					dat |= 1;
				}
				CAN_SPI_SCK_H;		// SCK パルス出力
				CAN_SPI_SCK_L;
			}
			data[p] = dat;
		}
	} else {
		// データ送信
		for(p = 0; p < cnt; p++) {
			dat = data[p];
			for(i = 0; i < 8; i++) {
				if(dat & 0x80) {
					// 1
					CAN_SPI_SO_H;
				} else {
					// 0
					CAN_SPI_SO_L;
				}
				dat <<= 1;
				CAN_SPI_SCK_H;		// SCK パルス出力
				CAN_SPI_SCK_L;
			}
		}
	}

	CAN_SPI_CS_H;

	return;
}


// --------------------------------------------------------------------
// CAN SID フィルタ、マスク設定 (兼用ルーチン)
// --------------------------------------------------------------------
void CANSetSidFilteMask(byte adrs, word sid) {
	CANWriteReg(adrs, (byte)(sid>>3));			// HIGH
	CANWriteReg(adrs + 1, (byte)(sid<<5));		// LOW
}

