//
// 第7章 CANバスモニタ用ノード Arduino版 ノード1
// 		copyright (c) 2008,2009,2010 T.Nakao
// 10/04/18
//

#include <stdio.h>
#include <ctype.h>

//#include "CANMsg.h"				// CANメッセージ
#include "CAN2515.h"			// CANドライバ

// シリアル出力マクロ
#define PUTSTR(s)	Serial.print(s)
#define PUTCH(c)	Serial.print(c)


// タイムスタンプ関係
byte Flag1Sec = false;
byte SubCount = 0;
byte Sec = 0;
byte Min = 0;

char StrBuf[16];		// シリアル出力用文字列バッファ
byte MsgBuf[14];		// CAN送受信バッファ ワーク

// シリアルコマンド関係
void SerCmdExec(void);
void CmdProc(char ch);

#define CMDBUF_SIZE 16
char CmdBuf[CMDBUF_SIZE];			// シリアルコマンドバッファ
byte CPtr = 0;				// シリアルコマンドバッファの挿入位置


// グローバル割り込み許可禁止
#define DIS_INT	noInterrupts()
#define ENA_INT	interrupts()

// タイマ1コンペアマッチ割り込みフラグ操作
#define CLR_TMR1_CMPA_IF	TIFR1|=(1<<OCF1A)			// 割り込みフラグクリア
#define ENA_TMR1_CMPA		TIMSK1 |= (1<<OCIE1A);		// 割り込み許可


//
// タイマ1 コンペアマッチ割り込み処理
//
ISR(TIMER1_COMPA_vect) {		// コンペアマッチ
//	CLR_TMR1_CMPA_IF;			// 自動でクリアされる

	SubCount++;
	if(SubCount == 10) {
//		Flag1Sec = true;
		SubCount = 0;
		Sec++;
		if(Sec == 60) {
			Sec = 0;
			Min++;
		}
	}
}


//
// 初期化
//
void setup() {
	Serial.begin(19200);

	// CAN関係の初期化
	CANInit(CAN_BRP_16MHz_125KBPS);	// CLOCK 16MHz CAN 125bps

	// フィルタ、マスク設定 SID 11bit
	CANSetSidFilter0(0x0000);
	CANSetSidFilter1(0x0000);
	CANSetSidFilter2(0x0000);
	CANSetSidFilter3(0x0000);
	CANSetSidFilter4(0x0000);
	CANSetSidFilter5(0x0000);
	CANSetSidMask0(0x0000);
	CANSetSidMask1(0x0000);

	CANSetFilterRxB0(3);			// RXB0でフィルタ、マスクは使用しない
	CANSetFilterRxB1(3);			// RXB1でフィルタ、マスクは使用しない

	// 受信バッファのロールオーバを許可
	CANBitModCmd(RXB0CTRL, 1<<BUKT, 1<<BUKT);
	CANSetOpMode(CAM_MODE_NORMAL);	// ノーマル・モード

	// MCP2515 SPI通信確認
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), 1<<B0BFE);		// LED3 OFF
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), (1<<B0BFS) | (1<<B0BFE));		// LED3 ON
	delay(1000);

	PUTSTR("\r\n** CANBUS MON **\r\n");

	// タイムスタンプ用変数クリア
	SubCount = 0;
	Sec = 0;
	Min = 0;

	// タイマ設定 100ms周期
	// 16MHz * 1/64 * 1/25000 = 10Hz
	// WGM13-10 = 0100, CS12-CS10 = 011
	// ClockSource 16MHz / 64, CTCモード
	TCCR1B |= ( (1<<WGM12) | (1<<CS11) | (1<<CS10) );
	TCCR1B &= ~(1<<CS12);
	TCCR1A &= ~((1<<WGM11) | (1<<WGM10));

	OCR1A = 25000;				// コンペア値

	TCNT1 = 0;					// タイマ値リセット
	CLR_TMR1_CMPA_IF;			// 割り込みフラグクリア

	ENA_TMR1_CMPA;				// TIMER1 コンペアマッチ割り込み許可
	ENA_INT;					// グローバル割り込み許可
}


//
// メインループ
//
void loop() {
	byte dat;

	byte reg;
	word msg;
	byte i, p;
	byte dlc;
	byte get_num = 0;

	// シリアル入力のチェック
	if(Serial.available() > 0) {
		dat = Serial.read();
		CmdProc(dat);
	}

	// ステータス・チェック
	reg = CANReadStat();
	if(reg & 1<<CANINTF_RX1IF) {
		// 受信あり ロールオーバ発生中
		MsgBuf[0] = CANReadReg(RXB1CTRL);

		// RXB1からメッセージ、データの読み出し
		CANRxB1Read(&MsgBuf[1], 8);
		get_num = 2;
	} else if(reg & 1<<CANINTF_RX0IF) {
		// 受信あり
		MsgBuf[0] = CANReadReg(RXB0CTRL);

		// RXB0からメッセージ、データの読み出し
		CANRxB0Read(&MsgBuf[1], 8);
		get_num = 1;
	}

	// データ受信チェック
	if(get_num != 0) {
		// 受信あり
		msg = ((word)MsgBuf[1] << 3);		// SIDH
		msg |= (MsgBuf[2] >> 5);			// SIDL

		sprintf(StrBuf, "%04X-", msg);		// メッセージ
		PUTSTR(StrBuf);

		reg = MsgBuf[2];				// DLC
		if(reg & (1<<SRR)) {
			// スタンダード・リモート・フレーム
			PUTCH('R');
		} else {
			// スタンダード・データ・フレーム
			PUTCH('D');
		}

		sprintf(StrBuf, "-%02dm%02d.%ds", Min, Sec, SubCount);
		PUTSTR(StrBuf);

		dlc = MsgBuf[5] & 0x0F;
		sprintf(StrBuf, "-%d ", dlc);
		PUTSTR(StrBuf);

		p = 6;
		for(i = 0; i < dlc; i++) {
			sprintf(StrBuf, "%02X ", MsgBuf[p++]);
			PUTSTR(StrBuf);
		}
		PUTCH('\r');
		PUTCH('\n');
	}
}


//
// シリアルコマンド処理
//
void SerCmdExec(void) {
	byte cmd, reg, i, change = false, cmdcmp = false;
	char ch;
//	byte p_cnf1;

	cmd = toupper(CmdBuf[0]);			// 1文字目を大文字に変換
	switch(cmd) {
		case 'B':				// ビットレート設定コマンド
			CANSetOpMode(CAM_MODE_CONFIG);	// コンフィギュレーション・モード

			// コンフィギュレーション・モードに切り替わったことを確認
			// 失敗時に10サイクルまで待つ
			for(i = 0; i < 10; i++) {
				reg = CANReadReg(CANSTAT);
				if((reg >> 5) == CAM_MODE_CONFIG) {
					change = true;
					break;					// 切替完了
				}
			}
			if(change) {
				// 切替完了時
				ch = CmdBuf[1];
				if(ch == '2') {
					// 500kHz 16TQ, 16MHz
					CANInit(CAN_BRP_16MHz_500KBPS);
					strcpy(StrBuf, "500K ");
					PUTSTR(StrBuf);
				} else if(ch == '1') {
					// 250kHz 16TQ, 16MHz
					CANInit(CAN_BRP_16MHz_250KBPS);
					strcpy(StrBuf, "250K ");
					PUTSTR(StrBuf);
				} else {
					// 125kHz 16TQ, 16MHz
					CANInit(CAN_BRP_16MHz_125KBPS);
					strcpy(StrBuf, "125K ");
					PUTSTR(StrBuf);
				}
//				CANSetOpMode(CAM_MODE_LISTEN);	// リスン・オンリ・モード
				cmdcmp = true;
			}
			break;
		case 'S':				// モニタ開始コマンド
			DIS_INT;			// 割り込み禁止
			SubCount = 0;
			Sec = 0;
			Min = 0;
			TCNT1 = 0;			// タイマ値リセット
			CLR_TMR1_CMPA_IF;	// 割り込みフラグクリア
			ENA_INT;			// 割り込み許可
			CANSetOpMode(CAM_MODE_LISTEN);	// リスン・オンリ・モード
			cmdcmp = true;
			break;
		case 'E':				// モニタ終了コマンド
			CANSetOpMode(CAM_MODE_CONFIG);	// コンフィギュレーション・モード
			cmdcmp = true;
			break;
		default:
			break;
	}

	// コマンド実行結果をPCへ通知
	if(cmdcmp) {
		PUTCH('O');
		PUTCH('K');
		PUTCH('\r');
		PUTCH('\n');
	} else {
		PUTCH('?');
		PUTCH('\r');
		PUTCH('\n');
	}
}


//
// シリアルコマンド文字列ビルド
//
void CmdProc(char ch) {
	switch(ch) {
		case '\b':						// BS
			if(CPtr > 0) {
				CPtr--;					// バッファから1文字削除
				PUTSTR("\b \b");
			}
			break;
		case '\r':				// Enter(コマンド確定)
			if(CPtr < CMDBUF_SIZE) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = 0x00;	// NULLコード
				PUTSTR("\n\r");			// 改行
				SerCmdExec();			// コマンドを解析して実行
			}
			CPtr = 0;					// コマンドバッファをクリア
			break;
		case ' ':	// (SP)入力のキャンセル
			CPtr = 0;					// バッファを空にする
			PUTSTR("\n\r");				// 改行
			break;
		case '\n':
			break;
		default:
			PUTCH(ch);				// エコーバック
			if(CPtr < CMDBUF_SIZE-1) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = ch;		// コマンド文字を登録
				CPtr++;					// 次の登録位置(次回に備えて)
			} else {
				// コマンドバッファが満杯のとき
				CPtr = 0;				// バッファを空にする
				PUTSTR("?\n\r");			// 入力エラー
			}
			break;
	}
}
