//
// 第8章 CANブリッジ用ノード Arduino版 ノード1
// 		copyright (c) 2008,2009,2010 T.Nakao
// 10/04/18
//

#include <stdio.h>
#include <ctype.h>

//#include "CANMsg.h"				// CANメッセージ
#include "CAN2515.h"			// CANドライバ

// シリアル出力マクロ
#define PUTSTR(s)	Serial.print(s)
#define PUTCH(c)	Serial.print(c)


// タイムスタンプ関係
byte Flag1Sec = false;
byte SubCount = 0;
byte Sec = 0;
byte Min = 0;

char StrBuf[32];		// シリアル出力用文字列バッファ
byte MsgBuf[14];		// CAN送受信バッファ ワーク

// シリアルコマンド関係
void SerCmdExec(void);
void CmdProc(char ch);

#define CMDBUF_SIZE 16
char CmdBuf[CMDBUF_SIZE];			// シリアルコマンドバッファ
byte CPtr = 0;				// シリアルコマンドバッファの挿入位置

bool EchoBack = true;
byte TxPend0, TxPend1, TxPend2;

byte CmdSetMask(void);		// マスク設定
byte CmdSetFilter(void);	// フィルタ設定
byte CmdSetFMOption(void);	// フィルタオプション
byte CmdSetMsg(void);		// メッセージ設定
byte CmdSetDat(void);		// データ設定
byte CmdReadSts(void);		// ステータス・リード

unsigned int atox(char *s);

unsigned int toX(char ch);
unsigned int to4X(char *str);

//#define A_TO_X(a) atox(a)
#define A_TO_X(a) to4X(a)

//
// 初期化
//
void setup() {
	Serial.begin(19200);

	// CAN関係の初期化
	CANInit(CAN_BRP_16MHz_125KBPS);	// CLOCK 16MHz CAN 125bps

	// フィルタ、マスク設定 SID 11bit
	CANSetSidFilter0(0x0000);
	CANSetSidFilter1(0x0000);
	CANSetSidFilter2(0x0000);
	CANSetSidFilter3(0x0000);
	CANSetSidFilter4(0x0000);
	CANSetSidFilter5(0x0000);
	CANSetSidMask0(0x0000);
	CANSetSidMask1(0x0000);

	CANSetFilterRxB0(3);			// RXB0でフィルタ、マスクは使用しない
	CANSetFilterRxB1(3);			// RXB1でフィルタ、マスクは使用しない

	// 受信バッファのロールオーバを許可
	CANBitModCmd(RXB0CTRL, 1<<BUKT, 1<<BUKT);
	CANSetOpMode(CAM_MODE_NORMAL);	// ノーマル・モード

	// MCP2515 SPI通信確認
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), 1<<B0BFE);		// LED3 OFF
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), (1<<B0BFS) | (1<<B0BFE));		// LED3 ON
	delay(1000);

	PUTSTR("\r\n** CAN BRIDGE **\r\n");
}


//
// メインループ
//
void loop() {
	byte dat;

	byte reg;
	word msg;
	byte i, p;
	byte dlc;
	byte get_num = 0;

	// シリアル入力のチェック
	if(Serial.available() > 0) {
		dat = Serial.read();
		CmdProc(dat);
	}

	// ステータス・チェック
	reg = CANReadStat();
	if(reg & 1<<CANINTF_RX1IF) {
		// 受信あり ロールオーバ発生中
		MsgBuf[0] = CANReadReg(RXB1CTRL);

		// RXB1からメッセージ、データの読み出し
		CANRxB1Read(&MsgBuf[1], 8);
		get_num = 2;
	} else if(reg & 1<<CANINTF_RX0IF) {
		// 受信あり
		MsgBuf[0] = CANReadReg(RXB0CTRL);

		// RXB0からメッセージ、データの読み出し
		CANRxB0Read(&MsgBuf[1], 8);
		get_num = 1;
	}

	// データ受信チェック
	if(get_num != 0) {
		// 受信あり
		// メッセージ受信応答
		msg = ((word)MsgBuf[1] << 3);			// SIDH
		msg |= (MsgBuf[2] >> 5);				// SIDL
		sprintf(StrBuf, ".M=%03X", msg);	// メッセージ
		PUTSTR(StrBuf);

		reg = MsgBuf[2];				// DLC
		if(reg & (1<<SRR)) {
			// スタンダード・リモート・フレーム
			strcpy(StrBuf, "-R\r\n");
		} else {
			// スタンダード・データ・フレーム
			strcpy(StrBuf, "-D\r\n");
		}
		PUTSTR(StrBuf);

//		sprintf(StrBuf, "-%02dm%02d.%ds", Min, Sec, SubCount);
//		PUTSTR(StrBuf);

		dlc = MsgBuf[5] & 0x0F;

		// データ受信応答
		if(dlc > 0) {
			sprintf(StrBuf, ".D=%d", dlc);		// データ長
			p = 6;
			for(i = 0; i < dlc; i++) {
				sprintf(StrBuf, "%02X ", MsgBuf[p++]);
				PUTSTR(StrBuf);
			}
			PUTCH('\r');
			PUTCH('\n');
		}
	}

	if(TxPend0) {
		// 送信バッファ0の送信が完了していないとき
		reg = CANReadReg(TXB0CTRL);			// 送信バッファ0 CTRL
		if(!(reg & 1<<TXREQ)) {
			// 送信完了
			TxPend0 = false;
			strcpy(StrBuf, ".T=0\r\n");
			PUTSTR(StrBuf);
		} else {
			if(reg & 1<<MLOA) {
				// 調停負け
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_MLOA1;
				//strcpy(StrBuf, "Retry(00)\r\n");
				//PUTSTR(StrBuf);
			} else if(reg & 1<<TXERR) {
				// 送信エラー
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_TXERR1;
				//strcpy(StrBuf, "Retry(01)\r\n");
				//PUTSTR(StrBuf);
			}
		}
	}

	// 送信完了ステータスのチェック(送信バッファ1)
	if(TxPend1) {
		// 送信バッファ0の送信が完了していないとき
		reg = CANReadReg(TXB1CTRL);			// 送信バッファ1 CTRL
		if(!(reg & 1<<TXREQ)) {
			// 送信完了
			TxPend1 = false;
			strcpy(StrBuf, ".T=1\r\n");
			PUTSTR(StrBuf);
		} else {
			if(reg & 1<<MLOA) {
				// 調停負け
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB1CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB1CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_MLOA2;
				//strcpy(StrBuf, "Retry(10)\r\n");
				//PUTSTR(StrBuf);
			} else if(reg & 1<<TXERR) {
				// 送信エラー
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB1CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB1CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_TXERR2;
				//strcpy(StrBuf, "Retry(11)\r\n");
				//PUTSTR(StrBuf);
			}
		}
	}

	// 送信完了ステータスのチェック(送信バッファ2)
	if(TxPend2) {
		// 送信バッファ0の送信が完了していないとき
		reg = CANReadReg(TXB2CTRL);			// 送信バッファ2 CTRL
		if(!(reg & 1<<TXREQ)) {
			// 送信完了
			TxPend2 = false;
			strcpy(StrBuf, ".T=2\r\n");
			PUTSTR(StrBuf);
		} else {
			if(reg & 1<<MLOA) {
				// 調停負け
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB2CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB2CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_MLOA2;
				//strcpy(StrBuf, "Retry(20)\r\n");
				//PUTSTR(StrBuf);
			} else if(reg & 1<<TXERR) {
				// 送信エラー
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB2CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB2CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_TXERR2;
				//strcpy(StrBuf, "Retry(21)\r\n");
				//PUTSTR(StrBuf);
			}
		}
	}
}


//
// シリアルコマンド処理
//
void SerCmdExec(void) {
	byte cmd, reg, i, sts = true;
	char ch;

	cmd = toupper(CmdBuf[0]);			// 1文字目を大文字に変換
	switch(cmd) {
		case 'K':				// マスク設定
			sts = CmdSetMask();
			break;
		case 'F':				// フィルタ設定
			sts = CmdSetFilter();
			break;
		case 'O':				// フィルタオプション
			sts = CmdSetFMOption();
			break;
		case 'M':				// メッセージ設定
			sts = CmdSetMsg();
			break;
		case 'D':				// データ設定
			sts = CmdSetDat();
			break;
		case 'T':				// 送信要求発行
			switch(CmdBuf[1]) {
				case '0':			// TXB0
					reg = TXB0CTRL;
					TxPend0 = true;
					break;
				case '1':			// TXB1
					reg = TXB1CTRL;
					TxPend1 = true;
					break;
				case '2':			// TXB2
					reg = TXB2CTRL;
					TxPend2 = true;
					break;
				default:
					sts = false;
					goto brk1;
			}
			CANBitModCmd(reg, 1<<TXREQ, 1<<TXREQ);		// フラグセット
			if(EchoBack) {
				sprintf(StrBuf, "_TXREQ%c", CmdBuf[1]);
				PUTSTR(StrBuf);
			}
			break;
		case 'S':				// ステータス・リード
			sts = CmdReadSts();
			break;
		case 'E':				// エコーバック切替
			if(CmdBuf[1] == '0') {
				strcpy(StrBuf, "_Echo OFF OK\r\n");
				PUTSTR(StrBuf);
				EchoBack = false;
			} else {
				strcpy(StrBuf, "_Echo ON");
				PUTSTR(StrBuf);
				EchoBack = true;
			}
			break;
		default:
			sts = false;
			break;
	}

brk1:
	// コマンド実行結果をPCへ通知
	if(EchoBack) {
		if(sts) {
			strcpy(StrBuf, " OK\r\n");
		} else {
			strcpy(StrBuf, " ???\r\n");
		}
		PUTSTR(StrBuf);
	}
	return;
}


//
// シリアルコマンド文字列ビルド
//
void CmdProc(char ch) {
	switch(ch) {
		case '\b':						// BS
			if(CPtr > 0) {
				CPtr--;					// バッファから1文字削除
				PUTSTR("\b \b");
			}
			break;
		case '\r':				// Enter(コマンド確定)
			if(CPtr < CMDBUF_SIZE) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = 0x00;	// NULLコード
				PUTSTR("\n\r");			// 改行
				SerCmdExec();			// コマンドを解析して実行
			}
			CPtr = 0;					// コマンドバッファをクリア
			break;
		case ' ':	// (SP)入力のキャンセル
			CPtr = 0;					// バッファを空にする
			PUTSTR("\n\r");				// 改行
			break;
		case '\n':
			break;
		default:
			PUTCH(ch);				// エコーバック
			if(CPtr < CMDBUF_SIZE-1) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = ch;		// コマンド文字を登録
				CPtr++;					// 次の登録位置(次回に備えて)
			} else {
				// コマンドバッファが満杯のとき
				CPtr = 0;				// バッファを空にする
				PUTSTR("?\n\r");			// 入力エラー
			}
			break;
	}
}


//
// シリアルコマンド処理
// マスク設定
//
byte CmdSetMask(void) {
	word mask;
	char buf[4];
	byte reg, i, change = false;

	CANSetOpMode(CAM_MODE_CONFIG);		// コンフィギュレーション・モード
	// モード変更完了待ち
	for(i = 0; i < 10; i++) {
		reg = CANReadReg(CANSTAT) & 0xE0;
		if(reg == CAM_MODE_CONFIG << 5) {
			change = true;
			break;
		}
	}
	if(!change) {
		return false;
	}

	// フィルタ値取り出し
	buf[0] = CmdBuf[2];
	buf[1] = CmdBuf[3];
	buf[2] = CmdBuf[4];
	buf[3] = 0;
	mask = A_TO_X(buf);

	switch(CmdBuf[1]) {
		case '0':			// mask0
			CANSetSidMask0(mask);
			break;
		case '1':			// mask1
			CANSetSidMask1(mask);
			break;
		default:
			CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
			return false;
	}

	// 設定内容確認の表示
	if(EchoBack) {
		sprintf(StrBuf, "MSK%c=%03X", CmdBuf[1], mask);
		PUTSTR(StrBuf);
	}
	CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
	return true;
}


//
// シリアルコマンド処理
// フィルタ設定
//
byte CmdSetFilter(void) {
	word filter;
	char buf[4];
	byte reg, i, change = false;

	CANSetOpMode(CAM_MODE_CONFIG);		// コンフィギュレーション・モード
	// モード変更完了待ち
	for(i = 0; i < 10; i++) {
		reg = CANReadReg(CANSTAT) & 0xE0;
		if(reg == CAM_MODE_CONFIG << 5) {
			change = true;
			break;
		}
	}
	if(!change) {
		return false;
	}

	// フィルタ値取り出し
	buf[0] = CmdBuf[2];
	buf[1] = CmdBuf[3];
	buf[2] = CmdBuf[4];
	buf[3] = 0;
	filter = A_TO_X(buf);

	switch(CmdBuf[1]) {
		case '0':			// Filter0
			CANSetSidFilter0(filter);
			break;
		case '1':			// Filter1
			CANSetSidFilter1(filter);
			break;
		case '2':			// Filter2
			CANSetSidFilter2(filter);
			break;
		case '3':			// Filter2
			CANSetSidFilter3(filter);
			break;
		case '4':			// Filter2
			CANSetSidFilter4(filter);
			break;
		case '5':			// Filter2
			CANSetSidFilter5(filter);
			break;
		default:
			CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
			return false;
	}

	// 設定内容確認の表示
	if(EchoBack) {
		sprintf(StrBuf, "_FLT%c=%03X", CmdBuf[1], filter);
		PUTSTR(StrBuf);
	}
	CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
	return true;
}


//
// シリアルコマンド処理
// フィルタ設定
//
byte CmdSetFMOption(void) {
	byte mod;
	byte reg, i, change = false;

	CANSetOpMode(CAM_MODE_CONFIG);		// コンフィギュレーション・モード

	// モード変更完了待ち
	for(i = 0; i < 10; i++) {
		reg = CANReadReg(CANSTAT) & 0xE0;
		if(reg == CAM_MODE_CONFIG << 5) {
			change = true;
			break;
		}
	}
	if(!change) {
		return false;
	}

	mod = CmdBuf[2] - '0';

	if(mod > 3) {
		CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
		return false;
	}

	switch(CmdBuf[1]) {
		case '0':			// RXB0
			CANSetFilterRxB0(mod);
			break;
		case '1':			// RXB1CTRL
			CANSetFilterRxB1(mod);
			break;
		default:
			CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
			return false;
	}

	// 設定内容確認の表示
	if(EchoBack) {
		sprintf(StrBuf, "_RXB%cCTRL RXM=%d", CmdBuf[1], mod);
		PUTSTR(StrBuf);
	}
	CANSetOpMode(CAM_MODE_NORMAL);		// ノーマル・モード
	return true;
}


//
// シリアルコマンド処理
// メッセージ設定
//
byte CmdSetMsg(void) {
	word msg;
	char buf[4];

	// メッセージ(SID)取り出し
	buf[0] = CmdBuf[2];
	buf[1] = CmdBuf[3];
	buf[2] = CmdBuf[4];
	buf[3] = 0;
	msg = A_TO_X(buf);

	// SID
	MsgBuf[1] = (byte)(msg >> 3);	// SIDH
	MsgBuf[2] = (byte)(msg << 5);	// SIDL
	MsgBuf[3] = 0;					// EID8
	MsgBuf[4] = 0;					// EID0

	switch(CmdBuf[1]) {
		case '0':			// TXB0
			CANTxB0MsgWrite(&MsgBuf[1]);
			break;
		case '1':			// TXB1
			CANTxB1MsgWrite(&MsgBuf[1]);
			break;
		case '2':			// TXB2
			CANTxB2MsgWrite(&MsgBuf[1]);
			break;
		default:
			return false;
	}

	// 設定内容確認の表示
	if(EchoBack) {
		sprintf(StrBuf, "_TXB%c MSG=%03X", CmdBuf[1], msg);
		PUTSTR(StrBuf);
	}
	return true;
}


//
// シリアルコマンド処理
// データ設定
//
byte CmdSetDat(void) {
	word msg;
	char buf[4], ftp;
	byte txbn, i, p, mp;
	byte dc, dlc;

	dc = CmdBuf[2] - '0';		// データ長
	if(dc > 8) {
		// 範囲外
		return false;
	}

	// データの取り出し
	buf[2] = 0;
	p = 4;
	mp = 6;
	for(i = 0; i < dc; i++) {
		buf[0] = CmdBuf[p++];
		buf[1] = CmdBuf[p++];
		MsgBuf[mp++] = A_TO_X(buf);
	}


	if(CmdBuf[3] == '0') {
		// データ・フレーム
		dlc = 0;				// RTR=0, DLC=0
		ftp = 'D';
	} else {
		// リモート・フレーム
		dlc = (1<<RTR);			// RTR=1, DLC=0
		ftp = 'R';
	}
	MsgBuf[5] = dlc | dc;		// データ長

	switch(CmdBuf[1]) {
		case '0':			// TXB0
			CANTxB0DatWrite(&MsgBuf[6]);
			txbn = 0;
			break;
		case '1':			// TXB1
			CANTxB1DatWrite(&MsgBuf[6]);
			txbn = 1;
			break;
		case '2':			// TXB2
			CANTxB2DatWrite(&MsgBuf[6]);
			txbn = 2;
			break;
		default:
			return false;
	}

	// 設定内容確認の表示
	if(EchoBack) {
		sprintf(StrBuf, "_TXB%c DLC=%d FRM=%c ", CmdBuf[1], dc, ftp);
		PUTSTR(StrBuf);
		mp = 6;
		for(i = 0; i < dc; i++) {
			sprintf(StrBuf, "%02X ", MsgBuf[mp++]);
			PUTSTR(StrBuf);
		}
	}
	return true;
}


//
// シリアルコマンド処理
// ステータス・リード
//
byte CmdReadSts(void) {
	byte reg;
	char str[10];

	switch(CmdBuf[1]) {
		case '0':		// SPIコマンド ステータス・リード
			reg = CANReadStat();
			strcpy(str, "STAT");
			break;
		case '1':		// SPIコマンド 受信ステータス・リード
			reg = CANReadRXStat();
			strcpy(str, "RXSTAT");
			break;
		case '2':		// CANINTF
			reg = CANReadReg(CANINTF);
			strcpy(str, "CANINTF");
			break;
		case '3':		// EFLG
			reg = CANReadReg(EFLG);
			strcpy(str, "EFLG");
			break;
		case '4':		// REC
			reg = CANReadReg(REC);
			strcpy(str, "REC");
			break;
		case '5':		// TEC
			reg = CANReadReg(TEC);
			strcpy(str, "TEC");
			break;
		default:
			return false;
	}

	if(EchoBack) {
		sprintf(StrBuf, "_%s=%02X\r\n", str, reg);
		PUTSTR(StrBuf);
	}

	// データの表示
	sprintf(StrBuf, ".S=%c%02X\r\n", CmdBuf[1], reg);
	PUTSTR(StrBuf);

	return true;
}


unsigned int atox(char *s)
{
	unsigned int rv=0;
	char sign=0;
	while(1) {
		unsigned char cv;
		cv = *s++;
		if((cv - '0') > 9) {
			cv |= 0x20;				// tolower
			if((cv - 'a') > 6) {
				break;
			} else {
				// A-F
				cv -= 'a';
				rv=rv*16+cv+10;
			}
		} else {
			// 0-9
			cv -= '0';
			rv=rv*16+cv;
		}
	}
	return rv;
}


unsigned int toX(char ch) {

	ch = toupper(ch);

	if( ('0' <= ch) && (ch <= '9') ) {
		// 0..9
		return ch - '0';		// 0-9
	} else if( ('A' <= ch) && (ch <= 'F') ) {
		// A..F
		return ch - 'A' + 10;
	}
	return 0;
}


unsigned int to4X(char *str) {
	unsigned int val = 0;

	while(*str != 0) {
		val *= 0x10;
		val += toX(*str++);
	}
	return val;
}


