//
// 第9章 PC I/F用ノード Arduino版 ノード1
// 		copyright (c) 2008,2009,2010 T.Nakao
// 10/04/17
//

#include <stdio.h>
#include <ctype.h>

//typedef unsigned int word;

#include "CANMsg.h"				// CANメッセージ
#include "CAN2515.h"			// CANドライバ

#define PUTSTR(s)	Serial.print(s)
#define PUTCH(c)	Serial.print(c)

char StrBuf[16];

byte MsgBuf[14];		// CAN送受信バッファ ワーク

byte TxPend0, TxPend1, TxPend2;

// シリアルコマンド関係
void SerCmdExec(void);

byte CmdTmpMeag(void);		//
byte CmdRelay(void);		//
byte CmdServo(void);		//
byte CmdPwm(void);			//
void CmdProc(char ch);
unsigned int atox(char *s);

// シリアルコマンド処理
#define CMDBUF_SIZE 16
char CmdBuf[CMDBUF_SIZE];			// シリアルコマンドバッファ
byte CPtr = 0;				// シリアルコマンドバッファの挿入位置


void setup() {
  Serial.begin(19200);


	// CAN関係の初期化
	CANInit(CAN_BRP_16MHz_125KBPS);	// CLOCK 16MHz CAN 125bps

	// フィルタ、マスク設定 SID 11bit
	CANSetSidFilter0(0x0100);		// RXB0 filter
	CANSetSidFilter1(0x0000);
	CANSetSidFilter2(0x0100);		// RXB1 filter
	CANSetSidFilter3(0x0000);
	CANSetSidFilter4(0x0000);
	CANSetSidFilter5(0x0000);

	CANSetSidMask0(0x0F00);			// RXB0 mask
	CANSetSidMask1(0x0F00);			// RXB1 mask

	CANSetFilterRxB0(1);			// RXB0でフィルタ、マスク使用
	CANSetFilterRxB1(1);			// RXB1でフィルタ、マスク使用

	// 受信バッファのロールオーバを許可
	CANBitModCmd(RXB0CTRL, 1<<BUKT, 1<<BUKT);
	CANSetOpMode(CAM_MODE_NORMAL);	// ノーマル・モード

	// MCP2515 SPI通信確認
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), 1<<B0BFE);		// LED3 OFF
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), (1<<B0BFS) | (1<<B0BFE));		// LED3 ON
	delay(1000);


	PUTSTR("CAN Bridge\r\n");


}

void loop() {
	byte dat;

	byte reg, dlc;
	word msg, tmp;
	byte i, p, ch;
	byte get_num = 0;


	//
	if(Serial.available() > 0) {
		dat = Serial.read();
		CmdProc(dat);
	}

	// ステータス・チェック
	reg = CANReadStat();
	if(reg & 1<<CANINTF_RX1IF) {
		// 受信あり ロールオーバ発生中
		MsgBuf[0] = CANReadReg(RXB1CTRL);

		// RXB1からメッセージ、データの読み出し
		CANRxB1Read(&MsgBuf[1], 8);
		get_num = 2;
	} else if(reg & 1<<CANINTF_RX0IF) {
		// 受信あり
		MsgBuf[0] = CANReadReg(RXB0CTRL);

		// RXB0からメッセージ、データの読み出し
		CANRxB0Read(&MsgBuf[1], 8);
		get_num = 1;
	}

	// データ受信チェック
	if(get_num != 0) {
		// 受信あり
		// メッセージを取り出す
		msg = ((word)MsgBuf[1] << 3);			// SIDH
		msg |= (MsgBuf[2] >> 5);				// SIDL

		dlc = MsgBuf[5] & 0x0F;

		switch(msg) {
			case MSG_N13_RSPTMP:		// 温度応答
				if(dlc == 2) {
					tmp = (word)MsgBuf[6] << 8;
					tmp |= MsgBuf[7];			// 摂氏温度の10倍値
					ch = MsgBuf[6] >> 4;
					sprintf(StrBuf, ".T=%03d\r\n", tmp);
					PUTSTR(StrBuf);				// PCへシリアル送信
				}
			case MSG_N12_SWSTS:			// node2 SWステータス
				if(dlc == 1) {
					sprintf(StrBuf, ".S=%d\r\n", MsgBuf[6]);
					PUTSTR(StrBuf);				// PCへシリアル送信
				}
				break;
		}
	}


	if(TxPend0) {
		// 送信バッファ0の送信が完了していないとき
		reg = CANReadReg(TXB0CTRL);			// 送信バッファ0 CTRL
		if(!(reg & 1<<TXREQ)) {
			// 送信完了
			TxPend0 = false;
		} else {
			if(reg & 1<<MLOA) {
				// 調停負け
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_MLOA1;
				//strcpy(StrBuf, "Retry(00)\r\n");
				//PUTSTR(StrBuf);
			} else if(reg & 1<<TXERR) {
				// 送信エラー
				// いったんTX0REQをリセットして送信をアボート
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 0);
				// 送信要求発行(再送)
				CANBitModCmd(TXB0CTRL, 1<<TXREQ, 1<<TXREQ);
//				SRLedDat |= SRL_TXERR1;
				//strcpy(StrBuf, "Retry(01)\r\n");
				//PUTSTR(StrBuf);
			}
		}
	}
}



//
// シリアルコマンド処理
//
void SerCmdExec(void) {
	byte cmd, sts = true;
	char ch;

	cmd = toupper(CmdBuf[0]);			// 1文字目を大文字に変換
	switch(cmd) {
		case 'M':				// 温度測定コマンド
			sts = CmdTmpMeag();
			break;
		case 'R':				// リレー制御コマンド
			sts = CmdRelay();
			break;
		case 'V':				// サーボ制御コマンド
			sts = CmdServo();
			break;
		case 'W':				// PWM制御コマンド
			sts = CmdPwm();
			break;
		default:
			sts = false;
			break;
	}

brk1:
	// コマンド実行結果をPCへ通知
//	if(EchoBack) {
		if(sts) {
			strcpy(StrBuf, " OK\r\n");
		} else {
			strcpy(StrBuf, " ???\r\n");
		}
		PUTSTR(StrBuf);
//	}
	return;
}


//
// 温度測定コマンド処理
//
byte CmdTmpMeag(void) {
	struct tag_TXBnCTRL *pctrl;
	byte ctrl;

	word msg = MSG_N31_INQTMP;

	// SID
	MsgBuf[1] = (byte)(msg >> 3);	// SIDH
	MsgBuf[2] = (byte)(msg << 5);	// SIDL
	MsgBuf[3] = 0;					// EID8
	MsgBuf[4] = 0;					// EID0
	MsgBuf[5] = 1<<RTR;				// DLC=0, RTR=1

	CANTxB0MsgWrite(&MsgBuf[1]);	// ロード

	// 送信開始
	CANWriteReg(TXB0CTRL, 1<<TXREQ);	// SPIコマンド

	TxPend0 = true;					// 送信完了のチェックを開始

	return true;
}


//
// リレー制御コマンド処理
//
byte CmdRelay(void) {
	word msg;
	byte nod, rynum, sw;

	// メッセージ(SID)取り出し
	nod   = CmdBuf[1];			// ノード番号
	rynum = CmdBuf[2];			// リレー番号
	sw    = CmdBuf[3];			// スイッチフラグ

	if(nod == '3') {
		msg = MSG_N31_RLYSW;	// ノード3宛
	} else if(nod == '4') {
		msg = MSG_N41_RLYSW;	// ノード4宛
	} else {
		return false;			// パラメータエラー
	}

	// SID
	MsgBuf[1] = (byte)(msg >> 3);	// SIDH
	MsgBuf[2] = (byte)(msg << 5);	// SIDL
	MsgBuf[3] = 0;					// EID8
	MsgBuf[4] = 0;					// EID0
	MsgBuf[5] = 2;					// RTR=0, DLC=2
	MsgBuf[6] = rynum - '0';		// to int
	MsgBuf[7] = sw - '0';			// to int

	CANTxB0Write(&MsgBuf[1], 2);	// 送信バッファへロード

	// 送信開始
	CANWriteReg(TXB0CTRL, 1<<TXREQ);	// SPIコマンド

	TxPend0 = true;					// 送信完了のチェックを開始

	return true;
}


//
// サーボコマンド処理
//
byte CmdServo(void) {
	char buf[4];
	byte val, num;
	word msg;

	msg = MSG_N51_SERVO;

	if(CmdBuf[1] == '0') {		// サーボ番号
		num = 0;				// ch0
	} else {
		num = 1;				// ch1
	}

	buf[0] = CmdBuf[2];			// 設定値(H)
	buf[1] = CmdBuf[3];			// 設定値(L)
	buf[2] = 0;

	val = atox(buf);			// サーボ角度値

	// SID
	MsgBuf[1] = (byte)(msg >> 3);	// SIDH
	MsgBuf[2] = (byte)(msg << 5);	// SIDL
	MsgBuf[3] = 0;					// EID8
	MsgBuf[4] = 0;					// EID0
	MsgBuf[5] = 2;					// RTR=0, DLC=2

	// data
	MsgBuf[6] = num;				// ch
	MsgBuf[7] = val;				// 角度
	CANTxB0Write(&MsgBuf[1], 2);	// 送信バッファへロード

	// 送信開始
	CANWriteReg(TXB0CTRL, 1<<TXREQ);

	TxPend0 = true;					// 送信完了のチェックを開始

	return true;
}


//
// PWM制御コマンド処理
//
byte CmdPwm(void) {
	char buf[4];
	byte val;
	word msg;

	msg = MSG_N51_PWM;

	// CmdBuf[1] PWM番号は未使用

	buf[0] = CmdBuf[2];			// 設定値(H)
	buf[1] = CmdBuf[3];			// 設定値(L)
	buf[2] = 0;

	val = atox(buf);			// PWM値

	// SID
	MsgBuf[1] = (byte)(msg >> 3);	// SIDH
	MsgBuf[2] = (byte)(msg << 5);	// SIDL
	MsgBuf[3] = 0;					// EID8
	MsgBuf[4] = 0;					// EID0
	MsgBuf[5] = 1;					// RTR=0, DLC=2

	// data
	MsgBuf[6] = val;				// PWM値
	CANTxB0Write(&MsgBuf[1], 1);	// 送信バッファへロード

	// 送信開始
	CANWriteReg(TXB0CTRL, 1<<TXREQ);

	TxPend0 = true;					// 送信完了のチェックを開始

	return true;
}


void CmdProc(char ch) {
	switch(ch) {
		case '\b':						// BS
			if(CPtr > 0) {
				CPtr--;					// バッファから1文字削除
				PUTSTR("\b \b");
			}
			break;
		case '\r':				// Enter(コマンド確定)
			if(CPtr < CMDBUF_SIZE) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = 0x00;	// NULLコード
				PUTSTR("\n\r");			// 改行
				SerCmdExec();			// コマンドを解析して実行
			}
			CPtr = 0;					// コマンドバッファをクリア
			break;
		case ' ':	// (SP)入力のキャンセル
			CPtr = 0;					// バッファを空にする
			PUTSTR("\n\r");				// 改行
			break;
		case '\n':
			break;
		default:
			PUTCH(ch);				// エコーバック
			if(CPtr < CMDBUF_SIZE-1) {
				// まだコマンドバッファに登録できるとき
				CmdBuf[CPtr] = ch;		// コマンド文字を登録
				CPtr++;					// 次の登録位置(次回に備えて)
			} else {
				// コマンドバッファが満杯のとき
				CPtr = 0;				// バッファを空にする
				PUTSTR("?\n\r");			// 入力エラー
			}
			break;
	}
}


unsigned int atox(char *s)
{
	unsigned int rv=0;
	char sign=0;
	while(1) {
		unsigned char cv;
		cv = *s++;
		if((cv - '0') > 9) {
			cv |= 0x20;				// tolower
			if((cv - 'a') > 6) {
				break;
			} else {
				// A-F
				cv -= 'a';
				rv=rv*16+cv+10;
			}
		} else {
			// 0-9
			cv -= '0';
			rv=rv*16+cv;
		}
	}
	return rv;
}
