//
// N5(#258 for Arduino)
//
// 10/04/16
//		サーボ、PWM
//

#include <Servo.h>
#include <stdio.h>

typedef unsigned int word;

#include "CANMsg.h"				// CANメッセージ
#include "CAN2515.h"			// CANドライバ

// サーボのインスタンス定義
Servo Servo1;
Servo Servo2;


// ---------------------
// #258用定義
// ---------------------

// シフトレジスタ(HC595)操作
#define SR_LAT 12
#define SR_DAT 3
#define SR_CLK 2
#define SR_DAT_1 digitalWrite(SR_DAT, HIGH)
#define SR_DAT_0 digitalWrite(SR_DAT, LOW)
#define SR_LAT_H digitalWrite(SR_LAT, HIGH)
#define SR_LAT_L digitalWrite(SR_LAT, LOW)
#define SR_CLK_H digitalWrite(SR_CLK, HIGH)
#define SR_CLK_L digitalWrite(SR_CLK, LOW)

// I/Oピン定義
#define SERVO1_PIN	9		// サーボ1
#define SERVO2_PIN	10		// サーボ2
#define FAN_PIN 	6		// ファン(DCモータ)


byte MsgBuf[14];		// CAN送受信バッファ ワーク

byte SRBuf = 0;			// シフトレジスタ(HC595)バッファ


//
// Arduino初期化
//
void setup()
{
	int i;
	byte dat;

	// シリアルLED用シフトレジスタ初期化
	pinMode(SR_LAT, OUTPUT);      //
	pinMode(SR_DAT, OUTPUT);      //
	pinMode(SR_CLK, OUTPUT);      //
	SR_LAT_L;
	SetSR(0x00);

	// CAN関係の初期化
	CANInit(CAN_BRP_16MHz_125KBPS);	// CLOCK 16MHz CAN 125bps

	CANSetFilterRxB0(3);			// RXB0でフィルタ、マスクは使用しない
	CANSetFilterRxB1(3);			// RXB1でフィルタ、マスクは使用しない

	CANSetOpMode(CAM_MODE_NORMAL);	// ノーマル・モード

	// MCP2515 SPI通信確認
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), 1<<B0BFE);		// LED3 OFF
	delay(200);
	CANBitModCmd(BFPCTRL, (1<<B0BFS) | (1<<B0BFE), (1<<B0BFS) | (1<<B0BFE));		// LED3 ON
	delay(1000);

	// アクチュエータの接続先の定義
	Servo1.attach(SERVO1_PIN);		// サーボ
	Servo2.attach(SERVO2_PIN);		// サーボ
	pinMode(FAN_PIN, OUTPUT);		// ファン(DCモータ) PWM

	// 初期化
	Servo1.write(90);				// サーボ1 ニュートラル位置
	Servo2.write(90);				// サーボ2 ニュートラル位置
	analogWrite(FAN_PIN, 0);		// ファン停止
}


//
// Arduinoメインループ
//
void loop()
{
	word msg;
	byte ctrl, dlc, val1, val2;

	// データ受信チェック
	if(CANRxCheck(0)) {			// 受信チェック
		// 受信あり

		MsgBuf[0] = CANReadReg(RXB0CTRL);
		CANRxB0Read(&MsgBuf[1], 2);
		msg = MAKE_SID(MsgBuf[1], MsgBuf[2]);

		// 受信メッセージ判定
		switch(msg) {
			case MSG_N51_SERVO:		// サーボ制御
				val1 = MsgBuf[6];			// チャネル番号 0/1
				val2 = MsgBuf[7];			// 角度データ(0-180)

				if(val1 == 0) {
					// ch0
					Servo1.write(val2);
				} else {
					// ch1
					Servo2.write(val2);
				}
				break;
			case MSG_N51_PWM:		// PWM制御
				val1 = MsgBuf[6];			// PWM値 0-255
				analogWrite(FAN_PIN, val1);


				break;
		}
	}
}


//
// シフトレジスタ(HC595)出力設定
//
void SetSR(byte val) {
  byte i;

  SRBuf = val;
  for(i = 0; i < 8; i++) {
    if((val & 0x80) != 0) {
      SR_DAT_1;
    } else {
      SR_DAT_0;
    }
    val <<= 1;
    SR_CLK_H;		// シフトパルス
    SR_CLK_L;
  }
  SR_LAT_H;			// ラッチパルス
  SR_LAT_L;
}

